/*
 * cp40_mem.c --
 *
 *	Memory allocation routines for CP40.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-2006 Eric Thibaut.
 *
 *	This file is part of CP40soft.
 *
 *	CP40soft is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	CP40soft is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with CP40soft; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

#include <string.h>
#include <stdlib.h>

#include "cp40.h"

extern void *cp40_mem_alloc(size_t size);
extern void *cp40_mem_alloc_clear(size_t size);
extern void *cp40_mem_realloc(void *ptr, size_t size);
extern void  cp40_mem_free(void *ptr);

extern void  cp40_dbuf_init(cp40_dbuf_t *dbuf);
extern void  cp40_dbuf_free(cp40_dbuf_t *dbuf);
extern int   cp40_dbuf_resize(cp40_dbuf_t *dbuf, size_t size);
extern int   cp40_dbuf_append(cp40_dbuf_t *dbuf, void *data, size_t nbytes);
extern void  cp40_dbuf_clear(cp40_dbuf_t *dbuf);

/*---------------------------------------------------------------------------*/

static void alloc_error(size_t size);
static void realloc_error(size_t size);

void cp40_mem_free(void *ptr) { if (ptr) free(ptr); }

void *cp40_mem_alloc(size_t size)
{
  void *ptr = malloc(size);
  if (! ptr) alloc_error(size);
  return ptr;
}

void *cp40_mem_alloc_clear(size_t size)
{
  void *ptr = malloc(size);
  if (ptr) memset(ptr, 0, size);
  else alloc_error(size);
  return ptr;
}

void *cp40_mem_realloc(void *ptr, size_t size)
{
  if (ptr) {
    if (! (ptr = realloc(ptr, size))) realloc_error(size);
  } else {
    if (! (ptr = malloc(size))) alloc_error(size);
  }
  return ptr;
}

static void alloc_error(size_t size)
{
  cp40_sys_error("couldn't allocate %lu bytes", (unsigned long)size);
}

static void realloc_error(size_t size)
{
  cp40_sys_error("couldn't reallocate %lu bytes", (unsigned long)size);
}

void *cp40_alloc(unsigned int size, const char *what)
{
  void *ptr;
  if (! (ptr = malloc(size))) {
    if (what && strlen(what)<30) {
      cp40_error("insufficient memory for %s", what);
    } else {
      cp40_error("insufficient memory %u bytes", size);
    }
    return NULL;
  }
  return ptr;
}

void *cp40_alloc_clear(unsigned int size, const char *what)
{
  void *ptr;
  if ((ptr = cp40_alloc(size, what))) {
    memset(ptr, 0, size);
  }
  return ptr;
}

/*---------------------------------------------------------------------------*/
/* DYNAMIC BUFFERS */

void cp40_dbuf_init(cp40_dbuf_t *dbuf)
{
  dbuf->size= 0;
  dbuf->max_size= 0;
  dbuf->base= NULL;
}

void cp40_dbuf_free(cp40_dbuf_t *dbuf)
{
  if (dbuf) {
    void *tmp= dbuf->base;
    cp40_dbuf_init(dbuf);
    if (tmp) free(tmp);
  }
}

int cp40_dbuf_resize(cp40_dbuf_t *dbuf, size_t new_max_size)
{
  void *new, *old;

  /* Make size in multiple of CP40_DBUF_GRAIN. */
  new_max_size= CP40_ROUND_UP(new_max_size, CP40_DBUF_GRAIN);

  if (new_max_size != dbuf->max_size) {
    if (new_max_size < dbuf->size) {
      return cp40_error("cannot truncate dynamic buffer below its actual size");
    }

    /* Allocate new buffer and copy old contents into fresh buffer. */
    if (! (new= malloc(new_max_size))) {
      return cp40_sys_error("insufficient memory for dynamic buffer");
    }
    old= dbuf->base;
    if (dbuf->size) memcpy(new, old, dbuf->size);

    /* Take care of updating buffer address before size in case we
       get interrupted by some signal. */
    dbuf->base= new;
    dbuf->max_size= new_max_size;

    /* Free old buffer. */
    if (old) free(old);
  }
  return CP40_OK;
}

int cp40_dbuf_append(cp40_dbuf_t *dbuf, void *data, size_t nbytes)
{
  if (nbytes>0) {
    if (CP40_DBUF_AVAILABLE(dbuf) < nbytes
	&& cp40_dbuf_resize(dbuf, dbuf->size+nbytes) != CP40_OK) {
      return CP40_ERROR;
    }
    memcpy(CP40_DBUF_TAIL(dbuf), data, nbytes);
    dbuf->size+= nbytes;
  }
  return CP40_OK;
}

void cp40_dbuf_clear(cp40_dbuf_t *dbuf)
{
  if (dbuf->base && dbuf->size) {
    memset(dbuf->base, 0, dbuf->size);
  }
}

/*---------------------------------------------------------------------------*/
/* INDEXED 2D ARRAYS */

int cp40_alloc_2d(void *index_ptr, void *data, size_t elsize,
		  int imin, int imax, int jmin, int jmax)
{
  size_t k, ncols, nrows, pitch, nbytes;
  char **pntr, *addr;

  /* Check arguments. */
  if (elsize <= 0) return cp40_error("bad element size");
  if (jmin > jmax) return cp40_error("bad row bounds for 2-D array index");
  if (imin > imax) return cp40_error("bad column bounds for 2-D array index");

  /* Allocate memory. */
  nrows = imax - imin + 1;
  ncols = jmax - jmin + 1;
  pitch = ncols*elsize; /* bytes per row */
  pntr = malloc(nrows*sizeof(char *));
  if (data) {
    if (! pntr) {
      cp40_error("insufficient memory for 2D index");
    }
  } else {
    nbytes = nrows*pitch;
    if (! pntr || ! (data = malloc(nbytes))) {
      if (pntr) free(pntr);
      return cp40_error("insufficient memory for 2D array");
    }
    memset(data, 0, nbytes);
  }

  /* Instanciate row pointer index. */
  addr = (char *)data - jmin*elsize;
  for (k=0 ; k<nrows ; ++k) {
    pntr[k] = addr;
    addr += pitch;
  }
  *(char ***)index_ptr = pntr - imin;
  return CP40_OK;
}

void cp40_free_2d(void *index, int free_data, size_t elsize,
		  int imin, int imax, int jmin, int jmax)
{
  if (elsize>0 && imax >= imin && jmax >= jmin) {
    char **pntr = (char **)index + imin;
    if (free_data) free(pntr[0] + jmin*elsize);
    free(pntr);
  }
}

/*---------------------------------------------------------------------------*/
