/*
 * cp40_socket.c --
 *
 *	Implement socket for CP40 (largely based on similar code in Tcl).
 *
 * Copyright (c) 1995-1997 Sun Microsystems, Inc.
 * Copyright (c) 1997, 2006 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#include "cp40.h"

#ifndef HAVE_SOCKET
# define HAVE_SOCKET 0
#endif

#if HAVE_SOCKET
# include <sys/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <netdb.h>
# include <unistd.h>
# include <fcntl.h>
# include <errno.h>
# include <string.h>
#endif

/*---------------------------------------------------------------------------*/
/* PRIVATE DEFINITIONS FOR SOCKETS */

#define SOCKET_BACKLOG   10 /* Maximum allowable length of the queue for
			       pending connections, see listen(2).  If this is
			       larger than SOMAXCONN, the actual value used is
			       silently truncated to SOMAXCONN (5 for OS9). */
#define SOCKET_BUFSIZE 4096 /* Minimum size of socket internal buffers (only
			       used if SO_SNDBUF and SO_RCVBUF are defined). */

#if HAVE_SOCKET
static int init_socket_address(struct sockaddr_in *sockaddr_ptr,
			       const char *host, int port);
/*	This function initializes sockaddr structure SOCKADDR_PTR for a
	given host (HOST=NULL implies INADDR_ANY) and port number PORT.
	Returns 1 if the host was valid, 0 if the host could not be converted
	to an IP address. */
#endif

/*---------------------------------------------------------------------------*/

int cp40_create_socket(int port, const char *host, const char *myaddr,
		      int myport, int flags)
{
#if HAVE_SOCKET
  int value;                             /* Used to set/get socket options. */
  socklen_t length;
  struct sockaddr_in  sockaddr;          /* Socket address. */
  struct sockaddr_in  mysockaddr;        /* Socket address for client. */
  struct sockaddr_in *bindsockaddr_ptr;  /* Socket address for binding. */
  int async = flags & CP40_SOCK_ASYNC;   /* Asynchronous connect? */
  int server = flags & CP40_SOCK_SERVER; /* Server? */
  int sock = -1;                         /* Socket file descriptor number. */

  /* Initializes socket address(es) and figure out sockect address to bind. */
  if (! init_socket_address(&sockaddr, host, port)) {
    goto error;
  }
  if (server) {
    bindsockaddr_ptr= &sockaddr;
  } else if (myaddr != NULL || myport != 0) {
    if (! init_socket_address(&mysockaddr, myaddr, myport)) {
      goto error;
    }
    bindsockaddr_ptr= &mysockaddr;
  } else {
    bindsockaddr_ptr= NULL;
  }

  /* Create socket and set socket options (minimum size of socket buffers). */
  if ((sock= socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    goto error;
  }
#ifdef SO_SNDBUF
  length= sizeof value;
  getsockopt(sock, SOL_SOCKET, SO_SNDBUF, &value, &length);
  if (value < SOCKET_BUFSIZE) {
    value= SOCKET_BUFSIZE;
    setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &value, sizeof value);
  }
#endif
#ifdef SO_RCVBUF
  length= sizeof value;
  getsockopt(sock, SOL_SOCKET, SO_RCVBUF, &value, &length);
  if (value < SOCKET_BUFSIZE) {
    value= SOCKET_BUFSIZE;
    setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &value, sizeof value);
  }
#endif

  /* Eventually, bind to the specified port and set up to reuse addresses
     automatically. */
  if (bindsockaddr_ptr != NULL) {
#ifdef SO_REUSEADDR
    value= 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &value, sizeof value);
#endif
    if (bind(sock, (struct sockaddr *)bindsockaddr_ptr,
	     sizeof(struct sockaddr))) {
      goto error;
    }
  }

  if (server) {
    /* Server socket: start listening for connections. */
    if (listen(sock, SOCKET_BACKLOG) == -1) {
      goto error;
    }
  } else {
    /* Client socket: attempt to connect. The connect may fail at present with
       an EINPROGRESS but at a later time it will complete.  The caller will
       set up a file handler on the socket if she is interested in being
       informed when the connect completes. */
    if (async && fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK) < 0) {
      goto error;
    }
    if (connect(sock, (struct sockaddr *)&sockaddr, sizeof(sockaddr))) {
      if (async && errno == EINPROGRESS) {
	errno= 0;
      } else {
	goto error;
      }
    }
  }
  return sock;

 error:
  if (sock != -1) {
    close(sock);
  }
  cp40_sys_error("cannot create socket");
#else /* not HAVE_SOCKET */
  cp40_error("cannot create socket (compiled without socket support)");
#endif /* HAVE_SOCKET */
  return -1;	/* Error. */
}
/*---------------------------------------------------------------------------*/
#if HAVE_SOCKET
static int init_socket_address(struct sockaddr_in *sockaddr_ptr,
			       const char *host, int port)
{
  struct in_addr addr;		/* For 64/32 bit madness */

  memset(sockaddr_ptr, 0, sizeof(*sockaddr_ptr));
  sockaddr_ptr->sin_family= AF_INET;
  sockaddr_ptr->sin_port= htons((unsigned short)(port & 0xFFFF));
  if (host == NULL) {
    addr.s_addr= INADDR_ANY;
  } else {
    addr.s_addr= inet_addr(host);
    if (addr.s_addr == -1) {
#if HAVE_GETHOSTBYNAME /* Use gethostbyname() to get host inet address. */
      struct hostent *hostent;		/* Host database entry */

      hostent= gethostbyname(host);
      if (hostent != NULL) {
	memcpy(&addr, hostent->h_addr_list[0], hostent->h_length);
      } else {
#   ifdef EHOSTUNREACH
	errno= EHOSTUNREACH;
#   else
#	ifdef ENXIO
	errno= ENXIO;
#	endif
#   endif
	return 0;	/* error */
      }
#else /* Cannot use gethostbyname(). */
      return 0;	/* error */
#endif
    }
  }

  /* NOTE: On 64 bit machines the assignment below is rumored to not do the
     right thing. Please report errors related to this if you observe
     incorrect behavior on 64 bit machines such as DEC Alphas.  Should we
     modify this code to do an explicit memcpy?  */

  sockaddr_ptr->sin_addr.s_addr= addr.s_addr;
  return 1;	/* Success. */
}

#endif /* HAVE_SOCKET */
/*---------------------------------------------------------------------------*/
